@@ -28,6 +28,7 @@
 #include <string.h>
 
 #include <gdk/gdkx.h>
+#include <gdk/gdkkeysyms.h>
 #include <gtk/gtk.h>
 
 #include "gs-window.h"
@@ -51,9 +52,10 @@ struct GSWindowPrivate
         GdkRectangle geometry;
 
         guint      lock_enabled : 1;
-        guint      logout_enabled : 1;
         guint      user_switch_enabled : 1;
+        guint      logout_enabled : 1;
         guint64    logout_timeout;
+        char      *logout_command;
 
         GtkWidget *box;
         GtkWidget *socket;
@@ -65,17 +67,22 @@ struct GSWindowPrivate
         guint      dialog_unmap_signal_id;
         guint      dialog_response_signal_id;
 
+        guint      watchdog_timer_id;
+
         gint       pid;
         gint       watch_id;
         gint       dialog_response;
 
         GList     *key_events;
 
+        gdouble    last_x;
+        gdouble    last_y;
+
         GTimer    *timer;
 };
 
 enum {
-        UNBLANKED,
+        DISACTIVATED,
         DIALOG_UP,
         DIALOG_DOWN,
         LAST_SIGNAL
@@ -85,6 +92,7 @@ enum {
         PROP_0,
         PROP_LOCK_ENABLED,
         PROP_LOGOUT_ENABLED,
+        PROP_LOGOUT_COMMAND,
         PROP_LOGOUT_TIMEOUT,
         PROP_MONITOR
 };
@@ -229,6 +237,16 @@ gs_window_real_realize (GtkWidget *widget)
                           widget);
 }
 
+/* every so often we should raise the window in case
+   another window has somehow gotten on top */
+static gboolean
+watchdog_timer (GSWindow *window)
+{
+        gtk_window_present (GTK_WINDOW (window));
+
+        return TRUE;
+}
+
 static void
 gs_window_real_show (GtkWidget *widget)
 {
@@ -243,6 +261,12 @@ gs_window_real_show (GtkWidget *widget)
         if (window->priv->timer)
                 g_timer_destroy (window->priv->timer);
         window->priv->timer = g_timer_new ();
+
+        if (window->priv->watchdog_timer_id != 0)
+                g_source_remove (window->priv->watchdog_timer_id);
+        window->priv->watchdog_timer_id = g_timeout_add (30000,
+                                                         (GSourceFunc)watchdog_timer,
+                                                         window);
 }
 
 void
@@ -253,6 +277,20 @@ gs_window_show (GSWindow *window)
         gtk_widget_show (GTK_WIDGET (window));
 }
 
+static void
+gs_window_real_hide (GtkWidget *widget)
+{
+        GSWindow *window;
+
+        window = GS_WINDOW (widget);
+
+        if (window->priv->watchdog_timer_id != 0)
+                g_source_remove (window->priv->watchdog_timer_id);
+
+        if (GTK_WIDGET_CLASS (parent_class)->hide)
+                GTK_WIDGET_CLASS (parent_class)->hide (widget);
+}
+
 void
 gs_window_destroy (GSWindow *window)
 {
@@ -270,9 +308,9 @@ gs_window_get_gdk_window (GSWindow *window)
 }
 
 static gboolean
-emit_unblanked_idle (GSWindow *window)
+emit_disactivated_idle (GSWindow *window)
 {
-        g_signal_emit (window, signals [UNBLANKED], 0);
+        g_signal_emit (window, signals [DISACTIVATED], 0);
 
         return FALSE;
 }
@@ -319,6 +357,12 @@ spawn_on_window (GSWindow *window,
         if (g_getenv ("XAUTHLOCALHOSTNAME"))
                 g_ptr_array_add (env, g_strdup_printf ("XAUTHLOCALHOSTNAME=%s",
                                                        g_getenv ("XAUTHLOCALHOSTNAME")));
+        if (g_getenv ("LANG"))
+                g_ptr_array_add (env, g_strdup_printf ("LANG=%s",
+                                                       g_getenv ("LANG")));
+        if (g_getenv ("LANGUAGE"))
+                g_ptr_array_add (env, g_strdup_printf ("LANGUAGE=%s",
+                                                       g_getenv ("LANGUAGE")));
         g_ptr_array_add (env, NULL);
 
         result = gdk_spawn_on_screen_with_pipes (GTK_WINDOW (window)->screen,
@@ -338,8 +382,10 @@ spawn_on_window (GSWindow *window,
                 g_free (g_ptr_array_index (env, i));
         g_ptr_array_free (env, TRUE);
 
-        if (!result)
+        if (! result) {
+                g_strfreev (argv);
                 return FALSE;
+        }
 
         if (pid)
                 *pid = child_pid;
@@ -378,6 +424,7 @@ plug_removed (GtkWidget *widget,
 {
         gtk_widget_hide (window->priv->socket);
         gtk_container_remove (GTK_CONTAINER (window), GTK_WIDGET (window->priv->box));
+        window->priv->box = NULL;
 
         return TRUE;
 }
@@ -490,6 +537,38 @@ gs_window_dialog_finish (GSWindow *window)
         remove_key_events (window);
 }
 
+/* very rudimentary animation for indicating an auth failure */
+static void
+shake_dialog (GSWindow *window)
+{
+        int   i;
+        guint left;
+        guint right;
+
+        for (i = 0; i < 9; i++) {
+                if (i % 2 == 0) {
+                        left = 30;
+                        right = 0;
+                } else {
+                        left = 0;
+                        right = 30;
+                }
+
+                if (! window->priv->box)
+                        break;
+
+                gtk_alignment_set_padding (GTK_ALIGNMENT (window->priv->box),
+                                           0, 0,
+                                           left,
+                                           right);
+
+                while (gtk_events_pending ())
+                        gtk_main_iteration ();
+
+                g_usleep (10000);
+        }
+}
+
 static gboolean
 command_watch (GIOChannel   *source,
                GIOCondition  condition,
@@ -516,6 +595,10 @@ command_watch (GIOChannel   *source,
                                 if (1 == sscanf (line, " WINDOW ID= 0x%x %c", &id, &c)) {
                                         create_socket (window, id);
                                 }
+                        } else if (strstr (line, "NOTICE=")) {
+                                if (strstr (line, "NOTICE=AUTH FAILED")) {
+                                        shake_dialog (window);
+                                }
                         } else if (strstr (line, "RESPONSE=")) {
                                 if (strstr (line, "RESPONSE=OK")) {
                                         window->priv->dialog_response = DIALOG_RESPONSE_OK;
@@ -546,13 +629,17 @@ command_watch (GIOChannel   *source,
                 gs_window_dialog_finish (window);
 
                 if (window->priv->dialog_response == DIALOG_RESPONSE_OK) {
-                        g_idle_add ((GSourceFunc)emit_unblanked_idle, window);
+                        g_idle_add ((GSourceFunc)emit_disactivated_idle, window);
                 }
 
                 gs_window_clear (window);
                 set_invisible_cursor (GTK_WIDGET (window)->window, TRUE);
                 g_signal_emit (window, signals [DIALOG_DOWN], 0);
 
+                /* reset the pointer positions */
+                window->priv->last_x = -1;
+                window->priv->last_y = -1;
+
                 window->priv->watch_id = 0;
                 return FALSE;
         }
@@ -568,6 +655,9 @@ is_logout_enabled (GSWindow *window)
         if (! window->priv->logout_enabled)
                 return FALSE;
 
+        if (! window->priv->logout_command)
+                return FALSE;
+
         elapsed = g_timer_elapsed (window->priv->timer, NULL);
 
         if (window->priv->logout_timeout < (elapsed * 1000))
@@ -595,6 +685,7 @@ popup_dialog_idle (GSWindow *window)
 
         if (is_logout_enabled (window)) {
                 command = g_string_append (command, " --enable-logout");
+                g_string_append_printf (command, " --logout-command='%s'", window->priv->logout_command);
         }
         if (is_user_switch_enabled (window)) {
                 command = g_string_append (command, " --enable-switch");
@@ -628,7 +719,7 @@ gs_window_request_unlock (GSWindow *window)
                 return;
 
         if (! window->priv->lock_enabled) {
-                g_idle_add ((GSourceFunc)emit_unblanked_idle, window);
+                g_idle_add ((GSourceFunc)emit_disactivated_idle, window);
                 return;
         }
 
@@ -701,6 +792,21 @@ gs_window_set_logout_timeout (GSWindow *window,
                 window->priv->logout_timeout = logout_timeout;
 }
 
+void
+gs_window_set_logout_command (GSWindow   *window,
+                              const char *command)
+{
+        g_return_if_fail (GS_IS_WINDOW (window));
+
+        g_free (window->priv->logout_command);
+
+        if (command) {
+                window->priv->logout_command = g_strdup (command);
+        } else {
+                window->priv->logout_command = NULL;
+        }
+}
+
 void
 gs_window_set_monitor (GSWindow *window,
                        int       monitor)
@@ -742,6 +848,9 @@ gs_window_set_property (GObject            *object,
         case PROP_LOGOUT_ENABLED:
                 gs_window_set_logout_enabled (self, g_value_get_boolean (value));
                 break;
+        case PROP_LOGOUT_COMMAND:
+                gs_window_set_logout_command (self, g_value_get_string (value));
+                break;
         case PROP_LOGOUT_TIMEOUT:
                 gs_window_set_logout_timeout (self, g_value_get_long (value));
                 break;
@@ -771,6 +880,9 @@ gs_window_get_property (GObject    *object,
         case PROP_LOGOUT_ENABLED:
                 g_value_set_boolean (value, self->priv->logout_enabled);
                 break;
+        case PROP_LOGOUT_COMMAND:
+                g_value_set_string (value, self->priv->logout_command);
+                break;
         case PROP_LOGOUT_TIMEOUT:
                 g_value_set_long (value, self->priv->logout_timeout);
                 break;
@@ -797,6 +909,13 @@ static void
 queue_key_event (GSWindow    *window,
                  GdkEventKey *event)
 {
+        /* Eat the first return or space */
+        if (window->priv->key_events == NULL
+            && (event->keyval == GDK_Return
+                || event->keyval == GDK_space)) {
+                return;
+        }
+
         window->priv->key_events = g_list_prepend (window->priv->key_events,
                                                    gdk_event_copy ((GdkEvent *)event));
 }
@@ -837,11 +956,32 @@ static gboolean
 gs_window_real_motion_notify_event (GtkWidget      *widget,
                                     GdkEventMotion *event)
 {
+        GSWindow *window;
+        gdouble   distance;
+        gdouble   min_distance = 10;
 
-        /* if we don't already have a socket then request an unlock */
-        if (! GS_WINDOW (widget)->priv->socket
-            && (GS_WINDOW (widget)->priv->request_unlock_idle_id == 0)) {
-                GS_WINDOW (widget)->priv->request_unlock_idle_id = g_idle_add ((GSourceFunc)gs_window_request_unlock_idle, widget);
+        window = GS_WINDOW (widget);
+
+        /* if the last position was not set then don't detect motion */
+        if (window->priv->last_x < 0 || window->priv->last_y < 0) {
+                window->priv->last_x = event->x;
+                window->priv->last_y = event->y;
+
+                return FALSE;
+        }
+
+        /* just an approximate distance */
+        distance = MAX (ABS (window->priv->last_x - event->x),
+                        ABS (window->priv->last_y - event->y));
+
+        if (distance > min_distance) {
+                /* if we don't already have a socket then request an unlock */
+                if (! window->priv->socket
+                    && (window->priv->request_unlock_idle_id == 0)) {
+                        window->priv->request_unlock_idle_id = g_idle_add ((GSourceFunc)gs_window_request_unlock_idle, widget);
+                }
+                window->priv->last_x = -1;
+                window->priv->last_y = -1;
         }
 
         return FALSE;
@@ -897,6 +1037,7 @@ gs_window_class_init (GSWindowClass *klass)
         object_class->set_property = gs_window_set_property;
 
         widget_class->show                = gs_window_real_show;
+        widget_class->hide                = gs_window_real_hide;
         widget_class->realize             = gs_window_real_realize;
         widget_class->key_press_event     = gs_window_real_key_press_event;
         widget_class->motion_notify_event = gs_window_real_motion_notify_event;
@@ -904,11 +1045,11 @@ gs_window_class_init (GSWindowClass *klass)
 
         g_type_class_add_private (klass, sizeof (GSWindowPrivate));
 
-        signals [UNBLANKED] =
-                g_signal_new ("unblanked",
+        signals [DISACTIVATED] =
+                g_signal_new ("disactivated",
                               G_TYPE_FROM_CLASS (object_class),
                               G_SIGNAL_RUN_LAST,
-                              G_STRUCT_OFFSET (GSWindowClass, unblanked),
+                              G_STRUCT_OFFSET (GSWindowClass, disactivated),
                               NULL,
                               NULL,
                               g_cclosure_marshal_VOID__VOID,
@@ -942,6 +1083,30 @@ gs_window_class_init (GSWindowClass *klass)
                                                                NULL,
                                                                FALSE,
                                                                G_PARAM_READWRITE));
+        g_object_class_install_property (object_class,
+                                         PROP_LOGOUT_ENABLED,
+                                         g_param_spec_boolean ("logout-enabled",
+                                                               NULL,
+                                                               NULL,
+                                                               FALSE,
+                                                               G_PARAM_READWRITE));
+        g_object_class_install_property (object_class,
+                                         PROP_LOGOUT_TIMEOUT,
+                                         g_param_spec_long ("logout-timeout",
+                                                            NULL,
+                                                            NULL,
+                                                            -1,
+                                                            G_MAXLONG,
+                                                            0,
+                                                            G_PARAM_READWRITE));
+        g_object_class_install_property (object_class,
+                                         PROP_LOGOUT_TIMEOUT,
+                                         g_param_spec_string ("logout-command",
+                                                              NULL,
+                                                              NULL,
+                                                              NULL,
+                                                              G_PARAM_READWRITE));
+
         g_object_class_install_property (object_class,
                                          PROP_MONITOR,
                                          g_param_spec_int ("monitor",
@@ -964,6 +1129,9 @@ gs_window_init (GSWindow *window)
         window->priv->geometry.width  = -1;
         window->priv->geometry.height = -1;
 
+        window->priv->last_x = -1;
+        window->priv->last_y = -1;
+
         gtk_window_set_decorated (GTK_WINDOW (window), FALSE);
 
         gtk_window_set_skip_taskbar_hint (GTK_WINDOW (window), TRUE);
@@ -997,10 +1165,22 @@ gs_window_finalize (GObject *object)
 
         g_return_if_fail (window->priv != NULL);
 
-        if (window->priv->request_unlock_idle_id)
+        g_free (window->priv->logout_command);
+
+        if (window->priv->watchdog_timer_id != 0) {
+                g_source_remove (window->priv->watchdog_timer_id);
+                window->priv->watchdog_timer_id = 0;
+        }
+
+        if (window->priv->request_unlock_idle_id != 0) {
                 g_source_remove (window->priv->request_unlock_idle_id);
-        if (window->priv->popup_dialog_idle_id)
+                window->priv->request_unlock_idle_id = 0;
+        }
+
+        if (window->priv->popup_dialog_idle_id != 0) {
                 g_source_remove (window->priv->popup_dialog_idle_id);
+                window->priv->popup_dialog_idle_id = 0;
+        }
 
         if (window->priv->timer)
                 g_timer_destroy (window->priv->timer);
